#############################################################################
# A file to implement and test the functions that have
# been written until now.
# To test this file, run "python3 -m unittest test.test" in the terminal.
#############################################################################

from jaggpy.classes import Scenario
from jaggpy.bf_solver import BFSolver
from jaggpy.asp_solver import ASPSolver

scenario1 = Scenario()
RELATIVE_FILE_PATH = "./test/testfiles/scenario1.jagg"
scenario1.load_from_file(RELATIVE_FILE_PATH)

scenario2 = Scenario()
RELATIVE_FILE_PATH2 = "./test/testfiles/scenario2.jagg"
scenario2.load_from_file(RELATIVE_FILE_PATH2)

scenario3 = Scenario()
RELATIVE_FILE_PATH3 = "./test/testfiles/scenario3.jagg"
scenario3.load_from_file(RELATIVE_FILE_PATH3)

scenario4 = Scenario()
RELATIVE_FILE_PATH4 = "./test/testfiles/scenario4.jagg"
scenario4.load_from_file(RELATIVE_FILE_PATH4)

scenario5 = Scenario()
RELATIVE_FILE_PATH5 = "./test/testfiles/scenario5.jagg"
scenario5.load_from_file(RELATIVE_FILE_PATH5)

scenario6 = Scenario()
RELATIVE_FILE_PATH6 = "./test/testfiles/scenario6.jagg"
scenario6.load_from_file(RELATIVE_FILE_PATH6)

scenario7 = Scenario()
RELATIVE_FILE_PATH7 = "./test/testfiles/scenario7.jagg"
scenario7.load_from_file(RELATIVE_FILE_PATH7)

scenario8 = Scenario()
RELATIVE_FILE_PATH8 = "./test/testfiles/scenario8.jagg"
scenario8.load_from_file(RELATIVE_FILE_PATH8)

brutus = BFSolver()

brutus.enumerate_outcomes(scenario1, "kemeny")
brutus.enumerate_outcomes(scenario1, "slater")
brutus.enumerate_outcomes(scenario2, "kemeny")
brutus.enumerate_outcomes(scenario2, "slater")
brutus.enumerate_outcomes(scenario3, "kemeny")
brutus.enumerate_outcomes(scenario3, "slater")
brutus.enumerate_outcomes(scenario4, "kemeny")
brutus.enumerate_outcomes(scenario4, "slater")
brutus.enumerate_outcomes(scenario5, "kemeny")
brutus.enumerate_outcomes(scenario5, "slater")
brutus.enumerate_outcomes(scenario6, "kemeny")

asp = ASPSolver()

asp.enumerate_outcomes(scenario1, "leximax")
asp.enumerate_outcomes(scenario1, "young")
asp.enumerate_first_n_outcomes(scenario1, "young", True, n=5)


asp.enumerate_outcomes(scenario2, "kemeny")
asp.enumerate_outcomes(scenario2, "leximax")
asp.enumerate_outcomes(scenario2, "young")
asp.enumerate_outcomes(scenario2, "slater")
asp.enumerate_outcomes(scenario2, "majority")

asp.enumerate_outcomes(scenario3, "kemeny")
asp.enumerate_outcomes(scenario3, "slater")

asp.enumerate_outcomes(scenario4, "kemeny")
asp.enumerate_outcomes(scenario4, "slater")

asp.enumerate_outcomes(scenario5, "kemeny")
asp.enumerate_outcomes(scenario5, "slater")

asp.enumerate_outcomes(scenario5, "kemeny")
asp.enumerate_outcomes(scenario5, "slater")

brutus.enumerate_outcomes(scenario6, "kemeny")
asp.enumerate_outcomes(scenario6, "kemeny")
brutus.enumerate_outcomes(scenario6, "slater")
asp.enumerate_outcomes(scenario6, "slater")

asp.enumerate_outcomes(scenario8, "kemeny")
asp.enumerate_outcomes(scenario8, "kemeny")

scenario1.pretty_repr()



#####################################################################
## The simplest solver that uses the given rule to give back the
## outcome of the judgment aggregation.
#####################################################################

import copy
import math
from itertools import combinations
from nnf import Var, Or, And # pylint: disable=unused-import
from .classes import Solver
from .parser import Parser

class BFSolver(Solver):
    """A brute solver for Judgment Aggregation."""
    def solve(self, scenario, rule, verbose=False):
        """Given a scenario object and the name of a rule
        this function will yield a list with all the outcomes
        of the judgment aggregation. The rule should be given
        as a string and can be one of the following lowercase commands:
            - kemeny
            - maxhamming
            - slater
        Set verbose to true if you want the program to inform you that
        it is computing with the given rule.
        """
        # Translate the agenda so that the issues are represented
        # by their labels. That labels correspond to the correct issues
        # will be achieved by adding constraints.
        parser = Parser()
        translated_agenda = parser.translate_agenda(scenario.agenda)

        # Make a list of all the judgment sets that are consistent
        # with the output constraints.
        my_string = ""
        for conjunct in scenario.output_constraints:
            my_string += f'({conjunct}) & '
        for conjunct in translated_agenda:
            my_string += f'({conjunct}) & '

        # Update what variables appear in the scenario.
        all_variables = []
        for var in scenario.variables:
            all_variables.append(var)
        for label in scenario.agenda.keys():
            all_variables.append(f'l{label}')

        # Make sure that each variable gets an assignment.
        for var in all_variables:
            my_string += f"({var} | ~{var}) & "
        my_string = my_string[:-3]

        # Preprocess the string representing the scenario.
        var_prefix = "my_var_"
        my_string_preprocessed = my_string
        for var in all_variables:
            my_string_preprocessed = my_string_preprocessed.replace(var, var_prefix + var)
        for var in all_variables:
            exec(f"{var_prefix}{var} = Var('{var}')")
        output_contstraint = eval(my_string_preprocessed)

        # List all the models that are consistent with the output constraints.
        consistent_outcomes =  list(output_contstraint.models())

        # We determine the outcome with a helper function that corresponds to the chosen rule.
        # Kemeny rule
        if rule == "kemeny":
            if verbose:
                print("Computing outcome with the brute force solver and the Kemeny rule...")
            outcomes = self.solve_kemeny(scenario, consistent_outcomes)

        # MaxHamming rule
        elif rule == "maxhamming":
            if verbose:
                print("Computing outcome with the brute force solver and the MaxHamming rule...")
            outcomes =  self.solve_max_hamming(scenario, consistent_outcomes)

        # Slater rule
        elif rule == "slater":
            if verbose:
                print("Computing outcome with the brute force solver and the Slater rule...")
            outcomes = self.solve_slater(scenario, consistent_outcomes)
        
        else:
            raise Exception (f"{rule} is not a recognized aggregation rule.")

        # Clean outcomes to only contain issues
        for i, outcome in enumerate(outcomes):
            translated_outcomes = {}
            for formula in outcome.keys():
                if formula[0] == 'l':
                    label = int(formula[1])
                    translation = scenario.agenda[label]
                    translated_outcomes[translation] = outcome[formula]
            outcomes[i] = translated_outcomes

        # Remove duplicates from outcomes.
        outcomes = [dict(t) for t in {tuple(outcome.items()) for outcome in outcomes}]

        yield outcomes

    def support_number(self, agenda, profile):
        """The function support_number gets an agenda and profile and returns a dictionary, where each key
        is the label of a formula and each value is the number of times the issue corresponding to the 
        label is voted for."""
        support_count = dict()
        for formula in agenda.values():
            support_count[formula] = 0
        for judgement_set in profile:
            times_accepted = judgement_set[0]
            accepted_formula = judgement_set[1]
            for formula in accepted_formula:
                support_count[formula] += times_accepted
        return support_count

    def solve_kemeny(self, scenario, consistent_outcomes):
        """Helper function for the solve function. Given a scenario and a list of
        consistent outcomes, returns a list of the outcomes corresponding to the
        Kemeny rule."""
        # Keep track of the maximum agreement score and initiate list of outcomes.
        max_agreement = 0
        outcomes = []

        # Check agreement score for each outcome and update list of outcomes accordingly.
        for outcome in consistent_outcomes:
            agreement_score = 0
            # For each formula in the pre-agenda, check how many agents agree
            # with the outcome and update agreement score.
            for label in scenario.agenda.keys():
                support = self.support_number(scenario.agenda, scenario.profile)
                if outcome[f'l{label}']:
                    agreement_score += support[scenario.agenda[label]]
                else:
                    agreement_score += scenario.number_voters - support[scenario.agenda[label]]

            if agreement_score == max_agreement:
                outcomes.append(outcome)
            elif agreement_score > max_agreement:
                max_agreement = agreement_score
                outcomes = [outcome]
        return outcomes

    def solve_max_hamming(self, scenario, consistent_outcomes):
        """Helper function for the solve function. Given a scenario and a list of
        consistent outcomes, returns a list of the outcomes corresponding to the
        MaxHamming rule."""
        # Keep track of minimum maximal Hamming distance and initialise list of outcomes.
        minimum_mhd = math.inf
        outcomes = []

        # Find max Hamming difference for each outcome and update
        for outcome in consistent_outcomes:
            max_hd = 0
            # Check the Hamming distance from the outcome to each judgment set
            # and track the maximal distance.
            for jud_set in scenario.profile:
                # Check Hamming distance.
                ham_dist = 0
                for key in scenario.agenda.keys():
                    if outcome[f'l{key}'] and scenario.agenda[key] not in jud_set[1]:
                        ham_dist += 1
                    elif not outcome[f'l{key}'] and scenario.agenda[key] in jud_set[1]:
                        ham_dist += 1

                # Update the maximum HD for the outcome.
                if ham_dist > max_hd:
                    max_hd = ham_dist

            # Update outcome set to include only those outcomes with the minimum
            # maximum Hamming distance (thus far).
            if max_hd == minimum_mhd:
                outcomes.append(outcome)
            elif max_hd < minimum_mhd:
                minimum_mhd = max_hd
                outcomes = [outcome]
        return outcomes

    def solve_slater(self, scenario, consistent_outcomes):
        """Helper function for the solve function. Given a scenario and a list of
        consistent outcomes, returns a list of the outcomes corresponding to the
        Slater rule."""
        # Determine the set of formulas that has a majority vote, add a
        # formula if it is accepted and add it with 'neg ' in front of
        # it when it is rejected. The 'neg ' is merely a prefix to read
        # later.
        majority_number = scenario.number_voters / 2
        majority_set = []
        support = self.support_number(scenario.agenda, scenario.profile)
        for key in scenario.agenda.keys():
            formula = scenario.agenda[key]
            if support[formula] > majority_number:
                majority_set.append(f'l{key}')
            elif support[formula] == majority_number:
                pass
            else:
                negated_formula = f'neg l{key}'
                majority_set.append(negated_formula)

        # Make a list of potential subsets to see if there are consistent subsets of
        # the current 'size'.
        potential_subsets = []
        for i in range(len(majority_set), 0, -1):
            # For a given length give all subsets of the majority_set
            # of that given length.
            subsets = list(combinations(majority_set, i))
            for subset in subsets:
                temp_outcomes = copy.deepcopy(consistent_outcomes)
                to_remove = []
                # For each formula in the subset, remove all the models
                # that do not agree with it, thereby checking which models
                # agree with all formulas in the subset.
                for formula in subset:
                    if formula[0:4] == "neg ":
                        for j, value in enumerate(temp_outcomes):
                            if value[formula[4:]]:
                                to_remove.append(j)
                    else:
                        for j, value in enumerate(temp_outcomes):
                            if not value[formula]:
                                to_remove.append(j)
                # Keep all models that have not been put in the to_remove list.
                to_keep = [index for index in range(len(temp_outcomes)) if index not in to_remove]
                consistent_list = [temp_outcomes[index] for index in to_keep]

                # If some models were kept we add this subset as a candidate.
                if consistent_list != []:
                    potential_subsets.append(subset)
            # We break out of the (decreasing) loop when we have at least one consistent subset, for 
            # this/these subsets will be the largest consistent subset(s).
            if potential_subsets != []:
                break
        # Go over all maximal (consistent) subsets.
        outcomes = []
        for subset in potential_subsets:
            temp_outcomes = copy.deepcopy(consistent_outcomes)
            to_remove = []
            # For each formula remove all the models that do not agree.
            for formula in subset:
                if formula[0:4] == "neg ":
                    for j, value in enumerate(temp_outcomes):
                        if value[formula[4:]]:
                            to_remove.append(j)
                else:
                    for j, value in enumerate(temp_outcomes):
                        if not value[formula]:
                            to_remove.append(j)
            to_keep = [index for index in range(len(temp_outcomes)) if index not in to_remove]
            # Add all the extensions of the chosen subsets to the outcome.
            for index in to_keep:
                outcomes.append(temp_outcomes[index])
        return outcomes



#####################################################################
## The ASP solver that uses the given rule to give back the
## outcome of the judgment aggregation. We use clingo to find the
## correct answer sets.
#####################################################################

import textwrap
import clingo
from .classes import Solver
from .parser import Parser

class ASPSolver(Solver):
    """A solver that uses Answer Set Programming to compute outcomes."""
    def solve(self, scenario, rule, verbose=False):
        """Given a scenario object and the name of a rule
        this function will yield the outcomes
        of the judgment aggregation. Each outcome is yielded seperately.
        The rule should be given as a string and can be one of the
        following lowercase commands:
            - kemeny
            - leximax
            - young
            - slater
            - majority
        """
        parser = Parser()

        # Create a list of all variables in the scenario.
        all_variables = set()
        for var in scenario.variables:
            all_variables.add(var)

        # Add the scenario to the asp_program using the scenario argument.
        asp_program = textwrap.dedent("""% We first add the scenario to our ASP program.
        """)

        # Adding issues.
        asp_program += textwrap.dedent("""
        % Adding the labels that represent the issues.
        """)
        for key in scenario.agenda:
            asp_program += f"issue(l{key}).\n"
            all_variables.add(f"l{key}")

        # Adding voters and judgment sets.
        asp_program += textwrap.dedent("""
        % Adding voters and specifying how they voted.
        """)
        voter_count = 0
        for coalition in scenario.profile:
            for voter_index in range(1,coalition[0]+1):
                # Register new voter.
                voter = str(voter_count + voter_index)
                asp_program += f"voter({voter}).\n"
                # Register how they voted for each issue.
                for label in scenario.agenda:
                    if scenario.agenda[label] in coalition[1]:
                        asp_program += f"js({voter},l{label}).\n"
                    else:
                        asp_program += f"js({voter},-l{label}).\n"
            voter_count += coalition[0]

        # Adding input constraints.
        asp_program += "\n% Declare input constraints (in CNF)\n"
        total_input_constraints = ""

        # Compound separate constraints into one.
        for conjunct in scenario.input_constraints:
            total_input_constraints += f"{conjunct} & "

        # Add auxiliary input constraints that guarantee that labels
        # correspond to the right formulas.
        for constraint in parser.translate_agenda(scenario.agenda):
            total_input_constraints += f"({constraint}) & "
        total_ic = total_input_constraints[:-3]

        # Translate the constraint to CNF.
        cnf_object = parser.to_cnf(total_ic, all_variables)
        ic_cnf = cnf_object[0]
        all_variables = all_variables.union(cnf_object[1])

        # Adding the input constraint clauses to the program.
        conjuncts = ("".join(ic_cnf.split())).split("&")
        clause_number = 1
        for clause in conjuncts:
            prep_clause = "".join(clause.split("("))
            prep_clause = "".join(prep_clause.split(")"))
            conjunct = prep_clause.split("|")
            for string in conjunct:
                if string[0] == "(":
                    formula = string[1:]
                elif string[-1] == ")":
                    formula = string[:-1]
                else:
                    formula = string
                if formula[0] == "~":
                    asp_program += f'inputClause({clause_number}, -{formula[1:]}).\n'
                else:
                    asp_program += f'inputClause({clause_number}, {formula}).\n'
            clause_number += 1

        # Adding output constraints.
        asp_program += "\n% Declare output constraints (in CNF)\n"
        total_output_contstraints = ""
        
        # Compound separate constraints into one.
        for conjunct in scenario.output_constraints:
            total_output_contstraints += f"{conjunct} & "

        # Add auxiliary input constraints that guarantee
        # that labels correspond to the right formulas.
        for constraint in parser.translate_agenda(scenario.agenda):
            total_output_contstraints += f"({constraint}) & "
        total_oc = total_output_contstraints[:-3]

        # Translate the constraint to CNF.
        cnf_object = parser.to_cnf(total_oc, all_variables)
        oc_cnf = cnf_object[0]
        all_variables = all_variables.union(cnf_object[1])

        # Adding the output constraint clauses to the program.
        conjuncts = ("".join(oc_cnf.split())).split("&")
        clause_number = 1
        for clause in conjuncts:
            prep_clause = "".join(clause.split("("))
            prep_clause = "".join(prep_clause.split(")"))
            conjunct = prep_clause.split("|")
            for string in conjunct:
                if string[0] == "(":
                    formula = string[1:]
                elif string[-1] == ")":
                    formula = string[:-1]
                else:
                    formula = string
                if formula[0] == "~":
                    asp_program += f'outputClause({clause_number}, -{formula[1:]}).\n'
                else:
                    asp_program += f'outputClause({clause_number}, {formula}).\n'
            clause_number += 1

        # Declare variables.
        asp_program += '\n'
        for variable in all_variables:
            asp_program += f'variable({variable}).\n'

        # Add the consistency checks for the input and output constraints.
        asp_program += textwrap.dedent("""
        % Consistency check with respect to the input constraint
        agent(A) :- voter(A).
        lit(X;-X) :- issue(X).
        1 { js(A,X) ; js(A,-X) } 1 :- agent(A), variable(X).
        :- voter(A), inputClause(C,_), js(A,-L) : inputClause(C,L).

        % Consistency check of the collective outcome with respect
        % to the output constraint
        agent(col).
        :- agent(col), outputClause(C,_), js(col,-L) : outputClause(C,L).
        """)

        # Add the ASP code corresponding to the rule that is to be executed.
        if rule == "kemeny":
            if verbose:
                print("Computing outcome with ASP and the Kemeny rule...")
            asp_program += textwrap.dedent("""
            % Kemeny rule
            wgt(X,N) :- lit(X), N = #count { A : voter(A), js(A,X) }.
            #maximize { N@1,wgt(X,N) : wgt(X,N), js(col,X) }.
            """)

        elif rule == "leximax":
            if verbose:
                print("Computing outcome with ASP and the Leximax rule...")
            asp_program += textwrap.dedent("""
            % Leximax rule
            wgt(X,N) :- lit(X), N = #count { A : voter(A), js(A,X) }.
            #maximize { 1@N,wgt(X,N) : wgt(X,N), js(col,X) }.
            """)

        elif rule == "young":
            if verbose:
                print("Computing outcome with ASP and the Young rule...")
            asp_program += textwrap.dedent("""
            % Young rule
            in(A) ; out(A) :- voter(A).
            inwgt(X,N) :- lit(X), N = #count{ A : voter(A), in(A), js(A,X) }.
            inmaj(X) :- lit(X), inwgt(X,N), inwgt(-X,M), N > M.
            js(col,X) :- inmaj(X).
            #minimize { 1@1,out(A) : out(A) }.
            """)

        elif rule == "slater":
            if verbose:
                print("Computing outcome with ASP and the Slater rule...")
            asp_program += textwrap.dedent("""
            % Slater rule
            % determine the majority outcome
            pc(X,N) :- lit(X), N = #count { A : voter(A), js(A,X) }.
            maj(X) :- lit(X), pc(X,N), pc(-X,M), N > M.
            % maximize agreement with the majority outcome
            #minimize { 1@10,maj(X) : maj(X), js(col,-X) }.
            """)

        elif rule == "majority":
            if verbose:
                print("Computing outcome with ASP and the Majority rule...")
            asp_program += textwrap.dedent("""
            % Majority rule
            % require that the collective outcome agrees with all issues
            % that have strictly more support than their negation
            pc(X,N) :- lit(X), N = #count { A : voter(A), js(A,X) }.
            maj(X) :- lit(X), pc(X,N), pc(-X,M), N > M.
            js(col,X) :- maj(X).
            """)
        else:
            raise Exception (f"{rule} is not a recognized aggregation rule.")

        # Add the outcome predicate.
        asp_program += textwrap.dedent("""
            outcome(X) :- agent(col), js(col, X), issue(X).
            #show outcome/1.
                """)

        # Ground and solve the program.
        control = clingo.Control(arguments=["--project"])
        control.add("base", [], asp_program)
        control.ground([("base", [])])
        control.configuration.solve.models = 0
        control.configuration.solve.opt_mode = "optN"

        # Yield the results of the program.
        with control.solve(yield_=True) as handle:
            for model in handle:
                if model.optimality_proven:
                    outcome = dict()
                    for label in scenario.agenda:
                        for atom in model.symbols(shown=True):
                            outcome[scenario.agenda[label]] = False
                            if f'outcome(l{label})' == str(atom):
                                outcome[scenario.agenda[label]] = True
                                break
                    yield outcome



######################################################################
## Useful python classes for the scenarios and the possible solvers.
#################################################################

# A scenario class that will allow us to create a scenario object by loading 
# information from a .jagg file. A scenario has an agenda, input constraints,
# output constraints and a profile.

from abc import ABC, abstractmethod
from itertools import islice
from nnf import Var, Or, And  # pylint: disable=unused-import
from jaggpy.parser import Parser

class Scenario:
    """A Scenario object has the following properties:
            - variables: a list of occurring variables
            - agenda: a dictionary with numbers as keys and
                formulas as values
            - input_constraints: a list of input constraints
            - output_constraints: a list of output constraints
            - profile: a list of judgment sets
            - number_voters: an integer specifying the number of voters
            """
    def __init__(self):
        """A Scenario object has the following properties:
            - variables: a list of occurring variables
            - agenda: a dictionary with numbers as keys and
                formulas as values
            - input_constraints: a list of input constraints
            - output_constraints: a list of output constraints
            - profile: a list of judgment sets
            - number_voters: an integer specifying the number of voters
            """
        self.agenda = dict()
        self.input_constraints = []
        self.output_constraints = []
        self.profile = []
        self.variables = []
        self.number_voters = 0

    def load_from_file(self, path):
        """Load the scenario from a .jagg file given its path.
        The path should be a raw string, i.e. of the form r"path/to/file".
        The file should have the following format, with each element being on
        a new line:
            - var_1,..., var_n: list of all the variables
            - Number of Formulas: The number of formulas in the pre-agenda
            - X, Formula: The formula labeled by the number X
            - In, Formula: The input constraint labeled by the text "In"
            - Out, Formula: The output constraint labeled by the text "Out"
            - Number of voters, Number of distinct judgment sets
            - J, l1;...;ln: A list of the labels of the formulas
                that are accepted. The rest is rejected.
                This profile occurs J times. The formulas should be
                given by the times they are selected and seperated
                by a semicolon. For example, "4, 2;4;5".
        A formula can contain the following operators:
            - The OR operator |
            - The AND operator &
            - The NOT operator ~
            - The IMPLIES operator ->
        Parentheses can be omitted where clear from context. """
        # Read the file and split all lines
        conn = open(path)
        text = conn.read()
        lines = text.splitlines()
        conn.close()

        # Remove blank lines and comments from the lines
        lines = [line for line in lines if line != "" and line[0] != "#"]

        # Create a parser object for future use
        parser = Parser()

        # Add the variables to the scenario
        self.variables = lines[0].split(", ")

        # Add the formulas to the agenda dictionary using the given label
        number_of_formulass = int(lines[1])
        for i in range(2, number_of_formulass+2):
            current_line = lines[i].split(", ")
            label = int(current_line[0])
            formula = current_line[1]
            self.agenda[label] = parser.to_nnf(formula)

        # Add the input constraints to the list of constraints
        line_number = number_of_formulass+2
        while lines[line_number].split(", ")[0] == "In":
            formula = lines[line_number].split(", ")[1]
            if formula == "":
                first_var = self.variables[0]
                formula = f'({first_var} | ~{first_var})'
            self.input_constraints.append(parser.to_nnf(formula))
            line_number += 1

        # Check consistency of the input constraints
        my_string = ""
        for conjunct in self.input_constraints:
            my_string += f"({conjunct}) & "
        my_string = my_string[:-3]
        if not self.check_consistency(my_string):
            raise Exception ("The input constraints are inconsistent")

        # Add the output constraints to the list of constraints
        while lines[line_number].split(", ")[0] == "Out":
            formula = lines[line_number].split(", ")[1]
            if formula == "":
                first_var = self.variables[0]
                formula = f'({first_var} | ~{first_var})'
            self.output_constraints.append(parser.to_nnf(formula))
            line_number += 1

        # Check consistency of the output constraints
        my_string = ""
        for conjunct in self.output_constraints:
            my_string += f"({conjunct}) & "
        my_string = my_string[:-3]
        if not self.check_consistency(my_string):
            raise Exception ("The output constraints are inconsistent")

        # Add the number of voters to the scenario
        self.number_voters += int(lines[line_number].split(", ")[0])

        # Add the list of accepted formulas to the profile dictionary
        # for each of the judgment sets.
        number_of_js = int(lines[line_number].split(", ")[1])
        for i in range(line_number+1, line_number+number_of_js+1):
            current_line = lines[i].split(", ")
            # If the list of accepted formulas is not empty continue
            if current_line[1] != '':
                label = int(current_line[0])
                formula_labels = list(map(int, current_line[1].split(";")))
                accepted_formulas = []
                for formula_label in formula_labels:
                    accepted_formulas.append(self.agenda[formula_label])

                # Check consistency of the judgment set with respect
                # to the input constraint
                my_string = ""
                for conjunct in self.input_constraints:
                    my_string += f"({conjunct}) & "
                for j in range(1, len(self.agenda)+1):
                    if j not in formula_labels:
                        my_string += f"(~{self.agenda[j]}) & "
                    else:
                        my_string += f"({self.agenda[j]}) & "
                my_string = my_string[:-3]
                self.check_consistency(my_string)
                if not self.check_consistency(my_string):
                    raise Exception (f"The judgment set on line {i} is inconsistent"\
                        " with the input constraints.")

                # Add the judgment set to the scenario
                self.profile.append([label, accepted_formulas])
            else:
                # If the all issues are rejected, add the empty list
                self.profile.append([label, []])

    def check_consistency(self, sentence):
        """The function check_consistency should receive an NNF-formula as a
        string. It return a Boolean indicating whether the formula is consistent."""
        my_string = sentence

        # Use a prefix to prevent variable name collisions
        # Add this prefix to all variables in the string
        var_prefix = "my_var_"
        my_string_preprocessed = my_string
        for var in self.variables:
            my_string_preprocessed = my_string_preprocessed.replace(var, var_prefix + var)

        # Declare variables (with prefix) and parse the formula with the
        # variable prefixes added
        for var in self.variables:
            exec(f"{var_prefix}{var} = Var('{var}')")
        formula = eval(my_string_preprocessed)

        # Return whether the formula is consistent
        return formula.consistent()


    def pretty_repr(self):
        """Returns string that represents the scenario object in a readable way"""
        scenario_string = "Variables:"
        for variable in self.variables:
            scenario_string += f"\n{variable}"
        scenario_string += "\n\nSub-agenda (label, formula):"
        for key in self.agenda:
            scenario_string += f"\n{key}, {self.agenda[key]}"
        scenario_string += "\n\nInput constraints:"
        for constraint in self.input_constraints:
            scenario_string += f"\n{constraint}"
        scenario_string +=  "\n\nOutput constraints:"
        for constraint in self.output_constraints:
            scenario_string += f"\n{constraint}"
        scenario_string += "\n\nProfile (times selected, accepted formulas):"
        for judgment_set in self.profile:
            accepted = "("
            for variable in judgment_set[1]:
                if accepted == "(":
                    accepted += variable
                else:
                    accepted += ", " + variable
            accepted += ")"
            scenario_string += (f"\n{judgment_set[0]}, " + accepted)
        return scenario_string

    # A solver class with an enumerate_outcomes function that enumerates
    # all the outcomes given a scenario and an aggregation rule.
class Solver(ABC):
    """The abstract class for solvers. """
    @abstractmethod
    def solve(self, scenario, rule, verbose=False):
        """Given a scenario and an aggregation rule, yields a generator
        with the corresponding outcomes."""

    def enumerate_outcomes(self, scenario, rule, verbose=False):
        """Given a scenario and an aggregation rule, prints all
        corresponding outcomes."""
        for outcome in self.solve(scenario, rule, verbose):
            print(outcome)

    def enumerate_first_n_outcomes(self, scenario, rule, verbose=False, n=1):
        """Given a scenario, an aggregation rule and an integer n,
        prints the first n corresponding outcomes."""
        n_outcomes = islice(self.solve(scenario, rule, verbose), n)
        for outcome in n_outcomes:
            print(outcome)



#####################################################################
## Parser for sentences
## The function parse_sentence is heavily inspired by
## https://github.com/pyparsing/pyparsing/blob/master/examples/simpleBool.py#L16
## It has been altered for this application.
#####################################################################

from pyparsing import infixNotation, opAssoc, Keyword, Word, alphanums, ParserElement
from nnf import Var, Or, And # pylint: disable=unused-import

class Parser:
    """A parser class for parsing formulas."""
    def parse_sentence(self, sentence):
        """Given a formula as a string, parse it and return a list of
        lists with the [] representing brackets. So x1 | (x2 & ~x3)
        returns [x1, |, [x2, &, [~, x3]]] """
        # Create space between negations and atoms
        prep_sentence = sentence.split("~")
        prep_sentence = "~ ".join(prep_sentence)

        ParserElement.enablePackrat()

		# Define classes to be built at parse time, as each matching
		# expression type is parsed. Each class has a as_list method
		# that returns itself and its elements in list representation
		# (recursively).
        class BoolOperand:
            """The class BoolOperand"""
            def __init__(self, t):
                self.label = t[0]

            def __str__(self) -> str:
                return self.label

            __repr__ = __str__

            def as_list(self):
                return self.label

        class BoolNot:
            """The class for negations."""
            def __init__(self, t):
                self.arg = t[0][1]

            def __str__(self) -> str:
                return "~" + str(self.arg)

            __repr__ = __str__

            def as_list(self):
                return ["~", self.arg.as_list()]

        class BoolBinOp:
            """The parent class for binary operations."""
            repr_symbol: str = ""

            def __init__(self, t):
                self.args = t[0][0::2]

            def __str__(self) -> str:
                sep = " %s " % self.repr_symbol
                return "(" + sep.join(map(str, self.args)) + ")"


        class BoolAnd(BoolBinOp):
            """The class for conjunctions."""
            repr_symbol = "&"
            def as_list(self):
                """Returns list of conjuncts with '&' symbol."""
                result = []
                for argument in self.args:
                    result.append(argument.as_list())
                    result.append("&")
                result = result[:-1]
                return result

        class BoolOr(BoolBinOp):
            """The class disjunctions."""
            repr_symbol = "|"
            def as_list(self):
                """Return list of disjuncts with '|' symbol."""
                result = []
                for argument in self.args:
                    result.append(argument.as_list())
                    result.append("|")
                result = result[:-1]
                return result

        class BoolImplies(BoolBinOp):
            """The class for implications."""
            repr_symbol = "->"
            def as_list(self):
                """Return implications in list form with '->' symbol."""
                return [self.args[0].as_list(), "->", self.args[1].as_list()]

        # Define what the operator symbols mean
        NOT = Keyword("~")
        AND = Keyword("&")
        OR = Keyword("|")
        IMPLIES = Keyword("->")

        # Atoms can be alphanumerals
        bool_operand = Word(alphanums)
        bool_operand.setParseAction(BoolOperand).setName("bool_operand")

        bool_expr = infixNotation(
        bool_operand,
        [
            # Define precedence of operations
            (NOT, 1, opAssoc.RIGHT, BoolNot),
            (OR, 2, opAssoc.LEFT, BoolOr),
            (AND, 2, opAssoc.LEFT, BoolAnd),
            (IMPLIES, 2, opAssoc.LEFT, BoolImplies)
        ],
        )

        # Create a parse object
        parsed_sentence = bool_expr.parseString(prep_sentence)[0]

        # Return the parsed sentence as a list
        return parsed_sentence.as_list()

    def to_nnf(self, sentence):
        """Given a formula as a string, returns a string of the
        formula converted to NNF."""
        # Parse the sentence and then convert it to NNF
        parsed = self.parse_sentence(sentence)
        nnf = self.to_nnf_parsed(parsed)

        return nnf

    def to_nnf_parsed(self, sentence):
        """Helper function for the recursive proces of the to_nnf function.
        Given a formula, returns the formula partially converted
        to NNF. """
        # If the sentence is just a string, leave it be
        if isinstance(sentence, str):
            return sentence
        # If it is negated, push the negation through
        if sentence[0] == '~':
            if isinstance(sentence[1], str):
                return "".join(sentence)
            if sentence[1][1] == '&':
                return self.neg_and(sentence[1])
            if sentence[1][1] == '|':
                return self.neg_or(sentence[1])
            if sentence[1][1] == '->':
                return self.neg_implies(sentence[1])
            # Remove double negations
            if sentence[1][0] == '~':
                return self.to_nnf_parsed(sentence[1][1])
            raise Exception("There seems to be something wrong..")
        # If it is not negated, rewrite implication and
        # search for more negations
        operator = sentence[1]
        if operator == '->':
            left = self.to_nnf_parsed(['~', sentence[0]])
            right = self.to_nnf_parsed(sentence[2])
            return " ".join(['(', left, '|', right, ')'])
        result = ['(']
        for i, value in enumerate(sentence):
            if i % 2 == 0:
                element = self.to_nnf_parsed(value)
                result.append(element)
                result.append(operator)
        result = result[:-1]
        result.append(')')
        return " ".join(result)

    def neg_and(self, sentence):
        """Helper function for to_nnf_parsed. Given a negated conjunction,
        returns a disjunction with the disjuncts negated. """
        result = ['(']
        for i, value in enumerate(sentence):
            if i % 2 == 0:
                element = self.to_nnf_parsed(['~', value])
                result.append(element)
                result.append('|')
        result = result[:-1]
        result.append(')')
        return " ".join(result)

    def neg_or(self, sentence):
        """Helper function for to_nnf_parsed. Given a negated disjunction,
        returns a conjuncts with the conjuncts negated. """
        result = ['(']
        for i, value in enumerate(sentence):
            if i % 2 == 0:
                element = self.to_nnf_parsed(['~', value])
                result.append(element)
                result.append('&')
        result = result[:-1]
        result.append(')')
        return " ".join(result)

    def neg_implies(self, sentence):
        """Helper function for to_nnf_parsed. Given a negated implication,
        returns a disjunction with the antecedent negated."""
        # Negate the antecedent
        antecedent = self.to_nnf_parsed(['~', sentence[0]])
        consequent = self.to_nnf_parsed(sentence[2])

        # Return the negated implication as a conjunction
        return " ".join(['(', antecedent, '|', consequent, ')'])

    def to_cnf(self, sentence, variables):
        """
        Translate a sentence from NNF (as produced by the to_nnf function), to CNF.
        The function needs a formula/sentence as a string and a list of all
        variables occuring in the sentence. It returns the CNF formula and
        an updated list of all occurring variables.
        """
        my_string = sentence
        all_variables = set()

        # Use a prefix to prevent variable name collisions
        # Add this prefix to all variables in the string
        var_prefix = "my_var_"
        my_string_preprocessed = my_string
        for var in variables:
            my_string_preprocessed = my_string_preprocessed.replace(var, var_prefix + var)

        # Declare variables (with prefix) and parse the formula with the
        # variable prefixes added
        for var in variables:
            exec(f"{var_prefix}{var} = Var('{var}')")
            all_variables.add(var)
        formula = eval(my_string_preprocessed)

        # Translate the formula to CNF and update what variables
        # occur in the formula.
        formula = formula.to_CNF()
        for var in formula.vars():
            if not isinstance(var, str):
                all_variables.add('a' + str(var)[:4])
            else:
                all_variables.add(var)

        # Translate formula to string.
        list_of_conjuncts = []
        for conjunct in formula:
            conj = "( "
            disjunct_counter = 0
            for disjunct in conjunct:
                if str(disjunct)[-1] == '>':
                    disjunct = "a".join(str(disjunct)[:-1].split('<'))
                if disjunct_counter < len(conjunct) - 1:
                    conj = conj + str(disjunct) + " | "
                else:
                    conj = conj + str(disjunct) + " )"
                disjunct_counter += 1
            list_of_conjuncts.append(conj)
        if len(formula) == 1:
            return [list_of_conjuncts[0], all_variables]
        formula_str = "( "
        for i, value in enumerate(list_of_conjuncts):
            if i < len(list_of_conjuncts) - 1:
                formula_str = formula_str + value + " & "
            else:
                formula_str = formula_str + value + " )"
        return [formula_str, all_variables]

    def translate_agenda(self, agenda):
        """Given a (sub)-agenda returns a list of constraints. For each issue a
        constraint is added of the form (label -> formula) and (formula -> label).
        Both are made sure to be NNF formulas. """
        new_constraints = []
        for label in agenda.keys():
            label_var = f'l{label}'
            formula = agenda[label]

            # Add label -> formula
            new_constraints.append(f'~{label_var} | {formula}')

            # Add formula -> label
            neg_formula = self.to_nnf(f'~ ({formula})')
            new_constraints.append(f'{neg_formula} | {label_var}')

        return new_constraints
        